__all__ = ["GUILogHandler", "MainWindow"]

import logging
import os
import time
from typing import List

from PySide2.QtCore import (QCoreApplication, QEventLoop, QMutex, QSettings,
                            Qt, QThread, Signal)
from PySide2.QtGui import QIcon
from PySide2.QtWidgets import (QAbstractItemView, QAction, QDockWidget,
                               QMainWindow, QMessageBox, QTableWidget,
                               QTableWidgetItem)

from QGrain.models.SampleDataset import SampleDataset
from QGrain.resolvers.GUIResolver import GUIResolver
from QGrain.resolvers.MultiprocessingResolver import MultiProcessingResolver
from QGrain.ui.AboutWindow import AboutWindow
from QGrain.ui.AllDistributionCanvas import AllDistributionCanvas
from QGrain.ui.ControlPanel import ControlPanel
from QGrain.ui.DataManager import DataManager
from QGrain.ui.DistributionCanvas import DistributionCanvas
from QGrain.ui.LossCanvas import LossCanvas
from QGrain.ui.PCAPanel import PCAPanel
from QGrain.ui.RecordedDataTable import RecordedDataTable
from QGrain.ui.SettingWindow import SettingWindow
from QGrain.ui.TaskWindow import TaskWindow

QGRAIN_ROOT_PATH = os.path.dirname(os.path.dirname(__file__))

class GUILogHandler(logging.Handler):
    def __init__(self, target_widget):
        logging.Handler.__init__(self)
        self.target = target_widget
        self.__mutex = QMutex()

    def emit(self, record):
        self.__mutex.lock()
        if record.levelno < self.level:
            return
        self.target.show_message(record.msg % record.args)
        self.__mutex.unlock()


class MainWindow(QMainWindow):
    sigDataSelected = Signal(int)
    sigSetup = Signal()
    sigCleanup = Signal()
    logger = logging.getLogger("root.MainWindow")
    gui_logger = logging.getLogger("GUI")
    def __init__(self, is_dark=True):
        super().__init__()
        if is_dark:
            self.icon_folder = os.path.join(QGRAIN_ROOT_PATH, "settings", "icons", "dark")
        else:
            self.icon_folder = os.path.join(QGRAIN_ROOT_PATH, "settings", "icons", "light")
        self.init_ui(is_dark)
        self.status_bar = self.statusBar()
        self.gui_fitting_thread = QThread()
        self.gui_resolver = GUIResolver()
        # disable multi-thread for debug
        self.gui_resolver.moveToThread(self.gui_fitting_thread)
        self.multiprocessing_fitting_thread = QThread()
        self.multiprocessing_resolver = MultiProcessingResolver()
        self.multiprocessing_resolver.moveToThread(self.multiprocessing_fitting_thread)
        self.data_manager = DataManager(self)
        self.connect_all()
        self.msg_box = QMessageBox(self)
        self.msg_box.setWindowFlags(Qt.Drawer)
        self.reset_dock_layout()

    def show_message(self, message):
        self.status_bar.showMessage(message)

    def init_ui(self, is_dark: bool):
        # Menu
        self.file_menu = self.menuBar().addMenu(self.tr("File"))
        self.load_action = QAction(QIcon(os.path.join(self.icon_folder, "load.png")), self.tr("Load"), self)
        self.file_menu.addAction(self.load_action)
        self.save_action = QAction(QIcon(os.path.join(self.icon_folder, "save.png")), self.tr("Save"), self)
        self.file_menu.addAction(self.save_action)
        self.docks_menu = self.menuBar().addMenu(self.tr("Docks"))
        self.pca_panel_action = QAction(QIcon(os.path.join(self.icon_folder, "pca.png")), self.tr("PCA Panel"), self)
        self.docks_menu.addAction(self.pca_panel_action)
        self.loss_canvas_action = QAction(QIcon(os.path.join(self.icon_folder, "canvas.png")), self.tr("Loss Canvas"), self)
        self.docks_menu.addAction(self.loss_canvas_action)
        self.all_distribution_canvas_action = QAction(QIcon(os.path.join(self.icon_folder, "canvas.png")), self.tr("All Distribution Canvas"), self)
        self.docks_menu.addAction(self.all_distribution_canvas_action)
        self.distribution_canvas_action = QAction(QIcon(os.path.join(self.icon_folder, "canvas.png")), self.tr("Distribution Canvas"), self)
        self.docks_menu.addAction(self.distribution_canvas_action)
        self.control_panel_action = QAction(QIcon(os.path.join(self.icon_folder, "control.png")), self.tr("Control Panel"), self)
        self.docks_menu.addAction(self.control_panel_action)
        self.raw_data_table_action = QAction(QIcon(os.path.join(self.icon_folder, "raw_data_table.png")), self.tr("Raw Data Table"), self)
        self.docks_menu.addAction(self.raw_data_table_action)
        self.recorded_data_table_action = QAction(QIcon(os.path.join(self.icon_folder, "recorded_data_table.png")), self.tr("Recorded Data Table"), self)
        self.docks_menu.addAction(self.recorded_data_table_action)
        self.reset_docks_actions = QAction(QIcon(os.path.join(self.icon_folder, "reset.png")), self.tr("Reset"), self)
        self.docks_menu.addAction(self.reset_docks_actions)
        self.settings_action = self.menuBar().addAction(self.tr("Settings"))
        self.about_action = self.menuBar().addAction(self.tr("About"))

        self.setDockNestingEnabled(True)
        # PCA Panel
        self.pca_panel_dock = QDockWidget(self.tr("PCA Panel"))
        self.pca_panel = PCAPanel(is_dark=is_dark)
        self.pca_panel_dock.setWidget(self.pca_panel)
        self.pca_panel_dock.setObjectName("PCAPanelDock")
        # Loss Canvas
        self.loss_canvas_dock = QDockWidget(self.tr("Loss Canvas"))
        self.loss_canvas = LossCanvas(is_dark=is_dark)
        self.loss_canvas_dock.setWidget(self.loss_canvas)
        self.loss_canvas_dock.setObjectName("LossCanvasDock")
        # All Distribution Canvas
        self.all_distribution_canvas_dock = QDockWidget(self.tr("All Distribution Canvas"))
        self.all_distribution_canvas = AllDistributionCanvas(is_dark=is_dark)
        self.all_distribution_canvas_dock.setWidget(self.all_distribution_canvas)
        self.all_distribution_canvas_dock.setObjectName("AllDistributionCanvasDock")
        # Distribution Canvas
        self.distribution_canvas_dock = QDockWidget(self.tr("Distribution Canvas"))
        self.distribution_canvas = DistributionCanvas(is_dark=is_dark)
        self.distribution_canvas_dock.setWidget(self.distribution_canvas)
        self.distribution_canvas_dock.setObjectName("DistributionCanvasDock")

        # Control Panel
        self.control_panel_dock = QDockWidget(self.tr("Control Panel"))
        self.control_panel = ControlPanel()
        self.control_panel_dock.setWidget(self.control_panel)
        self.control_panel_dock.setObjectName("ControlPanelDock")

        # Raw Data Table
        self.raw_data_dock = QDockWidget(self.tr("Raw Data Table"))
        self.raw_data_table = QTableWidget()
        self.raw_data_table.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.raw_data_table.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.raw_data_table.setAlternatingRowColors(True)
        self.raw_data_dock.setWidget(self.raw_data_table)
        self.raw_data_dock.setObjectName("RawDataTableDock")

        # Recorded Data Table
        self.recorded_data_dock = QDockWidget(self.tr("Recorded Data Table"))
        self.recorded_data_table = RecordedDataTable()
        self.recorded_data_dock.setWidget(self.recorded_data_table)
        self.recorded_data_dock.setObjectName("RecordedDataTableDock")

        self.settings_window = SettingWindow(self)
        self.about_window = AboutWindow(self)
        self.task_window = TaskWindow(self)

    def connect_all(self):
        self.control_panel.sigDistributionTypeChanged.connect(self.gui_resolver.on_distribution_type_changed)
        self.control_panel.sigDistributionTypeChanged.connect(self.multiprocessing_resolver.on_distribution_type_changed)
        self.control_panel.sigComponentNumberChanged.connect(self.gui_resolver.on_component_number_changed)
        self.control_panel.sigComponentNumberChanged.connect(self.multiprocessing_resolver.on_component_number_changed)
        self.control_panel.sigComponentNumberChanged.connect(self.distribution_canvas.on_component_number_changed)
        self.control_panel.sigFocusSampleChanged.connect(self.data_manager.on_focus_sample_changed)
        self.control_panel.sigFocusSampleChanged.connect(self.on_focus_sample_changed)
        self.control_panel.sigObserveIterationChanged.connect(self.distribution_canvas.on_observe_iteration_changed)
        self.control_panel.sigObserveIterationChanged.connect(self.loss_canvas.on_observe_iteration_changed)
        self.control_panel.sigInheritParamsChanged.connect(self.gui_resolver.on_inherit_params_changed)
        self.control_panel.sigGUIResolverFittingStarted.connect(self.gui_resolver.try_fit)

        self.control_panel.sigDataSettingsChanged.connect(self.data_manager.on_settings_changed)
        self.control_panel.sigGUIResolverFittingCanceled.connect(self.on_task_canceled)
        self.control_panel.sigMultiProcessingFittingStarted.connect(self.multiprocessing_resolver.execute_tasks)
        # Connect directly
        self.control_panel.record_button.clicked.connect(self.data_manager.record_current_data)

        self.distribution_canvas.sigExpectedMeanValueChanged.connect(self.gui_resolver.on_excepted_mean_value_changed)

        self.data_manager.sigDataLoaded.connect(self.on_data_loaded)
        self.data_manager.sigDataLoaded.connect(self.control_panel.on_data_loaded)
        self.data_manager.sigDataLoaded.connect(self.multiprocessing_resolver.on_data_loaded)
        self.data_manager.sigDataLoaded.connect(self.pca_panel.on_data_loaded)
        self.data_manager.sigDataLoaded.connect(self.all_distribution_canvas.on_data_loaded)
        self.data_manager.sigTargetDataChanged.connect(self.gui_resolver.on_target_data_changed)
        self.data_manager.sigTargetDataChanged.connect(self.distribution_canvas.show_target_distribution)
        self.data_manager.sigDataRecorded.connect(self.recorded_data_table.on_data_recorded)
        self.data_manager.sigShowFittingResult.connect(self.distribution_canvas.show_fitting_result)
        self.data_manager.sigShowFittingResult.connect(self.loss_canvas.show_fitting_result)

        self.gui_resolver.sigFittingStarted.connect(self.control_panel.on_fitting_started)
        self.gui_resolver.sigFittingFinished.connect(self.control_panel.on_fitting_finished)
        self.gui_resolver.sigFittingSucceeded.connect(self.control_panel.on_fitting_suceeded)
        self.gui_resolver.sigFittingSucceeded.connect(self.data_manager.on_fitting_suceeded)
        self.gui_resolver.sigFittingFailed.connect(self.control_panel.on_fitting_failed)

        self.multiprocessing_resolver.sigTaskInitialized.connect(self.task_window.on_task_initialized)
        self.multiprocessing_resolver.sigTaskStateUpdated.connect(self.task_window.on_task_state_updated)
        self.multiprocessing_resolver.sigTaskFinished.connect(self.task_window.on_task_finished)
        self.multiprocessing_resolver.sigTaskFinished.connect(self.data_manager.on_multiprocessing_task_finished)

        self.raw_data_table.cellClicked.connect(self.on_data_item_clicked)
        self.recorded_data_table.sigRemoveRecords.connect(self.data_manager.remove_data)
        self.recorded_data_table.sigShowDistribution.connect(self.distribution_canvas.show_fitting_result)
        self.recorded_data_table.sigShowLoss.connect(self.loss_canvas.show_fitting_result)
        self.recorded_data_table.sigGenerateDistributionVideo.connect(self.distribution_canvas.generate_video)
        # Dock menu actions
        self.pca_panel_action.triggered.connect(self.show_pca_panel_dock)
        self.loss_canvas_action.triggered.connect(self.show_loss_canvas_dock)
        self.all_distribution_canvas_action.triggered.connect(self.show_all_distribution_canvas_dock)
        self.distribution_canvas_action.triggered.connect(self.show_distribution_canvas_dock)
        self.control_panel_action.triggered.connect(self.show_control_panel_dock)
        self.raw_data_table_action.triggered.connect(self.show_raw_data_dock)
        self.recorded_data_table_action.triggered.connect(self.show_recorded_data_dock)
        self.reset_docks_actions.triggered.connect(self.reset_dock_layout)
        self.settings_action.triggered.connect(self.settings_window.show)
        self.about_action.triggered.connect(self.about_window.show)

        self.load_action.triggered.connect(self.data_manager.load_data)
        self.save_action.triggered.connect(self.data_manager.save_data)
        self.sigDataSelected.connect(self.control_panel.on_data_selected)

        self.sigSetup.connect(self.control_panel.setup_all)
        self.sigSetup.connect(self.data_manager.setup_all)
        self.sigSetup.connect(self.settings_window.setup_all)
        self.sigSetup.connect(self.multiprocessing_resolver.setup_all)
        self.sigCleanup.connect(self.data_manager.cleanup_all)
        self.sigCleanup.connect(self.multiprocessing_resolver.cleanup_all)

        self.settings_window.data_setting.sigDataSettingChanged.connect(self.data_manager.on_settings_changed)
        self.settings_window.algorithm_setting.sigAlgorithmSettingChanged.connect(self.gui_resolver.on_algorithm_settings_changed)
        self.settings_window.algorithm_setting.sigAlgorithmSettingChanged.connect(self.multiprocessing_resolver.on_algorithm_settings_changed)

    def reset_dock_layout(self):
        self.pca_panel_dock.show()
        self.loss_canvas_dock.show()
        self.all_distribution_canvas_dock.show()
        self.distribution_canvas_dock.show()
        self.control_panel_dock.show()
        self.raw_data_dock.show()
        self.recorded_data_dock.show()

        self.pca_panel_dock.setFloating(False)
        self.loss_canvas_dock.setFloating(False)
        self.all_distribution_canvas_dock.setFloating(False)
        self.distribution_canvas_dock.setFloating(False)
        self.control_panel_dock.setFloating(False)
        self.raw_data_dock.setFloating(False)
        self.recorded_data_dock.setFloating(False)

        self.addDockWidget(Qt.DockWidgetArea.LeftDockWidgetArea, self.pca_panel_dock)
        self.addDockWidget(Qt.DockWidgetArea.RightDockWidgetArea, self.raw_data_dock)
        self.splitDockWidget(self.raw_data_dock, self.recorded_data_dock, Qt.Orientation.Horizontal)
        self.splitDockWidget(self.pca_panel_dock, self.control_panel_dock, Qt.Orientation.Vertical)
        self.tabifyDockWidget(self.pca_panel_dock, self.loss_canvas_dock)
        self.tabifyDockWidget(self.pca_panel_dock, self.all_distribution_canvas_dock)
        self.tabifyDockWidget(self.pca_panel_dock, self.distribution_canvas_dock)

        self.resizeDocks((self.pca_panel_dock, self.control_panel_dock), (self.height()*0.65, self.height()*0.35), Qt.Orientation.Vertical)
        self.resizeDocks((self.pca_panel_dock, self.control_panel_dock, self.raw_data_dock, self.recorded_data_dock), (self.width()*0.5, self.width()*0.5, self.width()*0.25, self.width()*0.25), Qt.Orientation.Horizontal)

    def show_pca_panel_dock(self):
        self.pca_panel_dock.show()

    def show_loss_canvas_dock(self):
        self.loss_canvas_dock.show()

    def show_all_distribution_canvas_dock(self):
        self.all_distribution_canvas_dock.show()

    def show_distribution_canvas_dock(self):
        self.distribution_canvas_dock.show()

    def show_control_panel_dock(self):
        self.control_panel_dock.show()

    def show_raw_data_dock(self):
        self.raw_data_dock.show()

    def show_recorded_data_dock(self):
        self.recorded_data_dock.show()

    def on_data_loaded(self, dataset: SampleDataset):
        nrows = dataset.data_count
        ncols = len(dataset.classes)
        self.raw_data_table.setRowCount(nrows)
        self.raw_data_table.setColumnCount(ncols)
        self.raw_data_table.setHorizontalHeaderLabels(["{0:.2f}".format(class_value) for class_value in dataset.classes])
        self.raw_data_table.setVerticalHeaderLabels([str(sample.name) for sample in dataset.samples])
        for i, sample in enumerate(dataset.samples):
            QCoreApplication.processEvents(QEventLoop.ExcludeUserInputEvents)
            for j, value in enumerate(sample.distribution):
                item = QTableWidgetItem("{0:.4f}".format(value))
                item.setTextAlignment(Qt.AlignCenter)
                self.raw_data_table.setItem(i, j, item)

        # resize to tight layout
        self.raw_data_table.resizeColumnsToContents()
        self.logger.info("Data was loaded, and has been update to the table.")

    def on_data_item_clicked(self, row, column):
        self.sigDataSelected.emit(row)
        self.logger.debug("The item at [%d] row in raw data table was selected.", row)

    def on_focus_sample_changed(self, index):
        self.raw_data_table.setCurrentCell(index, 0)

    def on_task_canceled(self):
        self.gui_resolver.cancel()

    def setup_all(self):
        # must call this again to make the layout as expected
        self.reset_dock_layout()
        # emit signal to let other object to setup
        self.gui_fitting_thread.start()
        self.multiprocessing_fitting_thread.start()
        # emit after the threads are started
        self.sigSetup.emit()

        settings = QSettings(os.path.join(QGRAIN_ROOT_PATH, "settings", "ui.ini"), QSettings.IniFormat)
        settings.beginGroup("MainWindow")
        geometry = settings.value("geometry")
        if geometry is not None:
            self.restoreGeometry(geometry)
        windowState = settings.value("windowState")
        if windowState is not None:
            self.restoreState(windowState)
        settings.endGroup()

    def closeEvent(self, e):
        # TODO: add task running check
        self.sigCleanup.emit()
        self.hide()

        settings = QSettings(os.path.join(QGRAIN_ROOT_PATH, "settings", "ui.ini"), QSettings.IniFormat)
        settings.beginGroup("MainWindow")
        settings.setValue("geometry", self.saveGeometry())
        settings.setValue("windowState", self.saveState())
        settings.endGroup()

        time.sleep(2)
        self.gui_fitting_thread.terminate()
        self.multiprocessing_fitting_thread.terminate()
        e.accept()



import weakref
from enum import Enum, unique
from threading import Lock
from typing import Callable, Dict, Iterable, List, Tuple

import numpy as np
from scipy.special import gamma

INFINITESIMAL = 1e-100
FRACTION_PARAM_NAME = "f"
NAME_KEY = "Name"
BOUNDS_KEY = "Bounds"
DEFAULT_VALUE_KEY = "Default"
LOCATION_KEY = "Location"
COMPONENT_INDEX_KEY = "ComponentIndex"
PARAM_INDEX_KEY = "ParamIndex"


@unique
class DistributionType(Enum):
    Normal = 0
    Weibull = 1
    GeneralWeibull = 2


def check_component_number(component_number: int):
    # Check the validity of `component_number`
    if type(component_number) != int:
        raise TypeError(component_number)
    elif component_number < 1:
        raise ValueError(component_number)

def get_param_count(distribution_type: DistributionType) -> int:
    if distribution_type == DistributionType.Normal:
        return 2
    elif distribution_type == DistributionType.Weibull:
        return 2
    elif distribution_type == DistributionType.GeneralWeibull:
        return 3
    else:
        raise NotImplementedError(distribution_type)

def get_param_names(distribution_type: DistributionType) -> Tuple[str]:
    if distribution_type == DistributionType.Normal:
        return ("mu", "sigma")
    elif distribution_type == DistributionType.Weibull:
        return ("beta", "eta")
    elif distribution_type == DistributionType.GeneralWeibull:
        return ("mu", "beta", "eta")
    else:
        raise NotImplementedError(distribution_type)

def get_base_func_name(distribution_type: DistributionType) -> str:
    if distribution_type == DistributionType.Normal:
        return "normal"
    elif distribution_type == DistributionType.Weibull:
        return "weibull"
    elif distribution_type == DistributionType.GeneralWeibull:
        return "gen_weibull"
    else:
        raise NotImplementedError(distribution_type)

def get_param_bounds(distribution_type: DistributionType) -> Tuple[Tuple[float, float]]:
    if distribution_type == DistributionType.Normal:
        return ((INFINITESIMAL, None), (INFINITESIMAL, None))
    elif distribution_type == DistributionType.Weibull:
        return ((INFINITESIMAL, None), (INFINITESIMAL, None))
    elif distribution_type == DistributionType.GeneralWeibull:
        return ((INFINITESIMAL, None), (INFINITESIMAL, None), (INFINITESIMAL, None))
    else:
        raise NotImplementedError(distribution_type)

# in order to obtain better performance,
# the params of components should be different
def get_param_defaults(distribution_type: DistributionType, component_number: int) -> Tuple[Tuple]:
    check_component_number(component_number)
    if distribution_type == DistributionType.Normal:
        return tuple(((i*10, 2+i) for i in range(1, component_number+1)))
    elif distribution_type == DistributionType.Weibull:
        return tuple(((10+i, (i+1)*15) for i in range(1, component_number+1)))
    elif distribution_type == DistributionType.GeneralWeibull:
        return tuple(((0, 2+i, i*10) for i in range(1, component_number+1)))
    else:
        raise NotImplementedError(distribution_type)

def get_params(distribution_type: DistributionType, component_number: int) -> List[Dict]:
    check_component_number(component_number)
    params = []
    param_count = get_param_count(distribution_type)
    param_names = get_param_names(distribution_type)
    param_bounds = get_param_bounds(distribution_type)
    param_defaults = get_param_defaults(distribution_type, component_number)
    # generate params for all components
    for component_index, component_defaults in enumerate(param_defaults):
        for param_index, name, bounds, defalut in zip(range(param_count), param_names, param_bounds, component_defaults):
            params.append({NAME_KEY: name+str(component_index+1), BOUNDS_KEY: bounds,
                           DEFAULT_VALUE_KEY: defalut, COMPONENT_INDEX_KEY: component_index,
                           PARAM_INDEX_KEY: param_index, LOCATION_KEY: component_index*param_count+param_index})
    # generate fractions for front n-1 components
    for component_index in range(component_number-1):
        # the fraction of each distribution
        params.append({NAME_KEY: FRACTION_PARAM_NAME+str(component_index+1), BOUNDS_KEY: (0, 1),
                       DEFAULT_VALUE_KEY: 1/component_number, COMPONENT_INDEX_KEY: component_index,
                       LOCATION_KEY: component_number*param_count + component_index})
    sort_params_by_location_in_place(params)
    return params

def sort_params_by_location_in_place(params: List[Dict]):
    params.sort(key=lambda element: element[LOCATION_KEY])

def get_bounds(params: List[Dict]) -> Tuple[Tuple]:
    bounds = []
    for param in params:
        bounds.append(param[BOUNDS_KEY])
    return tuple(bounds)

def get_constrains(component_number: int) -> Tuple[Dict]:
    if component_number == 1:
        return ()
    elif component_number > 1:
        return ({'type': 'ineq', 'fun': lambda args:  1 - np.sum(args[1-component_number:]) + INFINITESIMAL})
    else:
        raise ValueError(component_number)

def get_defaults(params: List[Dict]) -> Tuple[float]:
    defaults = []
    for param in params:
        defaults.append(param[DEFAULT_VALUE_KEY])
    return tuple(defaults)

def get_lambda_str(distribution_type: DistributionType, component_number:int) -> str:
    base_func_name = get_base_func_name(distribution_type)
    param_count = get_param_count(distribution_type)
    param_names = get_param_names(distribution_type)
    if component_number == 1:
        return "lambda x, {0}: {1}(x, {0})".format(", ".join(param_names), base_func_name)
    elif component_number > 1:
        parameter_list = ", ".join(["x"] + [name+str(i+1) for i in range(component_number) for name in param_names] + [FRACTION_PARAM_NAME+str(i+1) for i in range(component_number-1)])
        # " + " to connect each sub-function
        # the previous sub-function str list means the m-1 sub-functions with n params `fj * base_func(x, param_1_j, ..., param_i_j, ..., param_n_j)`
        # the last sub-function str which represents `(1-f_1-...-f_j-...-f_m-1) * base_func(x, param_1_j, ..., param_i_j, ..., param_n_j)`
        previous_format_str = "{0}{1}*{2}(x, " + ", ".join(["{"+str(i+3)+"}{1}" for i in range(param_count)]) + ")"
        previous_sub_func_strs = [previous_format_str.format(FRACTION_PARAM_NAME, i+1, base_func_name, *param_names) for i in range(component_number-1)]
        last_format_str = "({0})*{1}(x, " + ", ".join(["{"+str(i+3)+"}{2}" for i in range(param_count)]) + ")"
        last_sub_func_str = last_format_str.format("-".join(["1"]+["f{0}".format(i+1) for i in range(component_number-1)]), base_func_name, component_number, *param_names)
        expression = " + ".join(previous_sub_func_strs + [last_sub_func_str])
        lambda_string = "lambda {0}: {1}".format(parameter_list, expression)
        return lambda_string
    else:
        raise ValueError(component_number)

# prcess the raw params list to make it easy to use
def process_params(distribution_type: DistributionType, component_number: int, fitted_params: Iterable) -> Tuple[Tuple[Tuple, float]]:
    param_count = get_param_count(distribution_type)
    if component_number == 1:
        assert len(fitted_params) == param_count
        return ((tuple(fitted_params), 1.0),)
    elif component_number > 1:
        assert len(fitted_params) == (param_count+1) * component_number - 1
        expanded = list(fitted_params) + [1.0-sum(fitted_params[component_number*param_count:])]
        return tuple(((tuple(expanded[i*param_count:(i+1)*param_count]), expanded[component_number*param_count+i]) for i in range(component_number)))
    else:
        raise ValueError(component_number)

# the pdf function of Normal distribution
def normal(x, mu, sigma):
    if sigma <= 0.0:
        return np.zeros_like(x, dtype=np.float64)
    else:
        return 1/(sigma*np.sqrt(2*np.pi))*np.exp(-np.square(x-mu)/(2*np.square(sigma)))

def double_normal(x, mu1, sigma1, mu2, sigma2, f1):
    return f1 * normal(x, mu1, sigma1) + (1-f1) * normal(x, mu2, sigma2)

def triple_normal(x, mu1, sigma1, mu2, sigma2, mu3, sigma3, f1, f2):
    return f1 * normal(x, mu1, sigma1) + f2 * normal(x, mu2, sigma2) + (1-f1-f2) * normal(x, mu3, sigma3)

def quadruple_normal(x, mu1, sigma1, mu2, sigma2, mu3, sigma3, mu4, sigma4, f1, f2, f3):
    return f1 * normal(x, mu1, sigma1) + f2 * normal(x, mu2, sigma2) + f3 * normal(x, mu3, sigma3) + (1-f1-f2-f3) * normal(x, mu4, sigma4)

def normal_mean(mu, sigma):
    if sigma <= 0.0:
        return np.nan
    else:
        return mu

def normal_median(mu, sigma):
    if sigma <= 0.0:
        return np.nan
    else:
        return mu

def normal_mode(mu, sigma):
    if sigma <= 0.0:
        return np.nan
    else:
        return mu

def normal_standard_deviation(mu, sigma):
    if sigma <= 0.0:
        return np.nan
    else:
        return sigma

def normal_variance(mu, sigma):
    if sigma <= 0.0:
        return np.nan
    else:
        return sigma**2

def normal_skewness(mu, sigma):
    if sigma <= 0.0:
        return np.nan
    else:
        return 0.0

def normal_kurtosis(mu, sigma):
    if sigma <= 0.0:
        return np.nan
    else:
        return 0.0


# The pdf function of Weibull distribution
def weibull(x, beta, eta):
    results = np.zeros_like(x, dtype=np.float64)
    if beta <= 0.0 or eta <= 0.0:
        return results
    else:
        non_zero = np.greater(x, 0.0)
        results[non_zero] = (beta/eta) * (x[non_zero]/eta)**(beta-1) * np.exp(-(x[non_zero]/eta)**beta)
        return results
    # return (beta/eta) * (x/eta)**(beta-1) * np.exp(-(x/eta)**beta)

def double_weibull(x, beta1, eta1, beta2, eta2, f):
    return f * weibull(x, beta1, eta1) + (1-f) * weibull(x, beta2, eta2)

def triple_weibull(x, beta1, eta1, beta2, eta2, beta3, eta3, f1, f2):
    return f1 * weibull(x, beta1, eta1) + f2 * weibull(x, beta2, eta2) + (1-f1-f2) * weibull(x, beta3, eta3)

def quadruple_weibull(x, beta1, eta1, beta2, eta2, beta3, eta3, beta4, eta4, f1, f2, f3):
    return f1 * weibull(x, beta1, eta1) + f2 * weibull(x, beta2, eta2) + f3 * weibull(x, beta3, eta3) + (1-f1-f2-f3) * weibull(x, beta4, eta4)

def weibull_mean(beta, eta):
    if beta <= 0.0 or eta <= 0.0:
        return np.nan
    else:
        return eta*gamma(1/beta+1)

def weibull_median(beta, eta):
    if beta <= 0.0 or eta <= 0.0:
        return np.nan
    else:
        return eta*(np.log(2)**(1/beta))

def weibull_mode(beta, eta):
    if beta <= 0.0 or eta <= 0.0:
        return np.nan
    elif beta <= 1:
        return 0.0
    else:
        return eta*(1-1/beta)**(1/beta)

def weibull_standard_deviation(beta, eta):
    if beta <= 0.0 or eta <= 0.0:
        return np.nan
    else:
        return eta*np.sqrt(gamma(2/beta+1) - gamma(1/beta+1)**2)

def weibull_variance(beta, eta):
    if beta <= 0.0 or eta <= 0.0:
        return np.nan
    else:
        return (eta**2)*(gamma(2/beta+1)-gamma(1/beta+1)**2)

def weibull_skewness(beta, eta):
    if beta <= 0.0 or eta <= 0.0:
        return np.nan
    else:
        return (2*gamma(1/beta+1)**3 - 3*gamma(2/beta+1)*gamma(1/beta+1) + gamma(3/beta+1)) / (gamma(2/beta+1)-gamma(1/beta+1)**2)**(3/2)

def weibull_kurtosis(beta, eta):
    if beta <= 0.0 or eta <= 0.0:
        return np.nan
    else:
        return (-3*gamma(1/beta+1)**4 + 6*gamma(2/beta+1)*gamma(1/beta+1)**2 - 4*gamma(3/beta+1)*gamma(1/beta+1) + gamma(4/beta+1)) / (gamma(2/beta+1)-gamma(1/beta+1)**2)**2


def gen_weibull(x, mu, beta, eta):
    return weibull(x-mu, beta, eta)

def double_gen_weibull(x, mu1, beta1, eta1, mu2, beta2, eta2, f):
    return f * gen_weibull(x, mu1, beta1, eta1) + (1-f) * gen_weibull(x, mu2, beta2, eta2)

def triple_gen_weibull(x, mu1, beta1, eta1, mu2, beta2, eta2, mu3, beta3, eta3, f1, f2):
    return f1 * gen_weibull(x, mu1, beta1, eta1) + f2 * gen_weibull(x, mu2, beta2, eta2) + (1-f1-f2)*gen_weibull(x, mu3, beta3, eta3)

def quadruple_gen_weibull(x, mu1, beta1, eta1, mu2, beta2, eta2, mu3, beta3, eta3, mu4, beta4, eta4, f1, f2, f3):
    return f1 * gen_weibull(x, mu1, beta1, eta1) + f2 * gen_weibull(x, mu2, beta2, eta2) + f3 * gen_weibull(x, mu3, beta3, eta3) + (1-f1-f2-f3) * gen_weibull(x, mu4, beta4, eta4)

def gen_weibull_mean(mu, beta, eta):
    return weibull_mean(beta, eta) + mu

def gen_weibull_median(mu, beta, eta):
    return weibull_median(beta, eta) + mu

def gen_weibull_mode(mu, beta, eta):
    return weibull_mode(beta, eta) + mu

def gen_weibull_standard_deviation(mu, beta, eta):
    return weibull_standard_deviation(beta, eta)

def gen_weibull_variance(mu, beta, eta):
    return weibull_variance(beta, eta)

def gen_weibull_skewness(mu, beta, eta):
    return weibull_skewness(beta, eta)

def gen_weibull_kurtosis(mu, beta, eta):
    return weibull_kurtosis(beta, eta)


def get_single_func(distribution_type: DistributionType) -> Callable:
    if distribution_type == DistributionType.Normal:
        return normal
    elif distribution_type == DistributionType.Weibull:
        return weibull
    elif distribution_type == DistributionType.GeneralWeibull:
        return gen_weibull
    else:
        raise NotImplementedError(distribution_type)

def get_param_by_mean(distribution_type: DistributionType, component_number: int, mean_values: Iterable):
    assert len(mean_values) == component_number
    param_count = get_param_count(distribution_type)
    func_params = get_params(distribution_type, component_number)
    param_values = list(get_defaults(func_params))
    if distribution_type == DistributionType.Normal:
        for i in range(component_number):
            # for normal distribution
            # only change the loaction param (first param of each component)
            param_values[i*param_count] = mean_values[i]
    elif distribution_type == DistributionType.Weibull:
        for i in range(component_number):

            beta = param_values[i*param_count]
            param_values[i*param_count+1] = mean_values[i] / gamma(1/beta+1)
    elif distribution_type == DistributionType.GeneralWeibull:
        for i in range(component_number):
            mu = param_values[i*param_count]
            beta = param_values[i*param_count+1]
            param_values[i*param_count+2] = (mean_values[i]-mu) / gamma(1/beta+1)
    else:
        raise NotImplementedError(distribution_type)
    return tuple(param_values)


class AlgorithmData:
    __cache = weakref.WeakValueDictionary()
    __cache_lock = Lock()

    def __init__(self, distribution_type: DistributionType, component_number: int):
        check_component_number(component_number)
        self.__distribution_type = distribution_type
        self.__component_number = component_number
        self.__param_count = get_param_count(self.distribution_type)
        self.__param_names = get_param_names(self.distribution_type)
        self.__single_func = get_single_func(distribution_type)
        self.__lambda_str = get_lambda_str(distribution_type, component_number)
        self.__mixed_func = self.__get_func_by_lambda_str(self.__lambda_str)
        self.__func_params = get_params(distribution_type, component_number)
        self.__bounds = get_bounds(self.__func_params)
        self.__defaults = get_defaults(self.__func_params)
        self.__constrains = get_constrains(component_number)
        self.__get_statistic_func()

    def __get_func_by_lambda_str(self, lambda_str: str) -> Callable:
        local_params = {"__tempMixedFunc": None}
        exec("__tempMixedFunc=" + lambda_str, None, local_params)
        mixed_func = local_params["__tempMixedFunc"]
        return mixed_func

    def __get_statistic_func(self):
        if self.distribution_type == DistributionType.Normal:
            self.__mean = normal_mean
            self.__median = normal_median
            self.__mode = normal_mode
            self.__standard_deviation = normal_standard_deviation
            self.__variance = normal_variance
            self.__skewness = normal_skewness
            self.__kurtosis = normal_kurtosis
        elif self.distribution_type == DistributionType.Weibull:
            self.__mean = weibull_mean
            self.__median = weibull_median
            self.__mode = weibull_mode
            self.__standard_deviation = weibull_standard_deviation
            self.__variance = weibull_variance
            self.__skewness = weibull_skewness
            self.__kurtosis = weibull_kurtosis
        elif self.distribution_type == DistributionType.GeneralWeibull:
            self.__mean = gen_weibull_mean
            self.__median = gen_weibull_median
            self.__mode = gen_weibull_mode
            self.__standard_deviation = gen_weibull_standard_deviation
            self.__variance = gen_weibull_variance
            self.__skewness = gen_weibull_skewness
            self.__kurtosis = gen_weibull_kurtosis
        else:
            raise NotImplementedError(self.distribution_type)

    @property
    def distribution_type(self) -> DistributionType:
        return self.__distribution_type

    @property
    def component_number(self) -> int:
        return self.__component_number

    @property
    def param_count(self) -> int:
        return self.__param_count

    @property
    def param_names(self) -> Tuple[str]:
        return self.__param_names

    @property
    def single_func(self) -> Callable:
        return self.__single_func

    @property
    def mixed_func(self) -> Callable:
        return self.__mixed_func

    @property
    def bounds(self) -> Tuple[Tuple]:
        return self.__bounds

    @property
    def defaults(self) -> Tuple[float]:
        return self.__defaults

    @property
    def constrains(self) -> Tuple[Dict]:
        return self.__constrains

    @property
    def mean(self) -> Callable:
        return self.__mean

    @property
    def median(self) -> Callable:
        return self.__median

    @property
    def mode(self) -> Callable:
        return self.__mode

    @property
    def variance(self) -> Callable:
        return self.__variance

    @property
    def standard_deviation(self) -> Callable:
        return self.__standard_deviation

    @property
    def skewness(self) -> Callable:
        return self.__skewness

    @property
    def kurtosis(self) -> Callable:
        return self.__kurtosis

    @classmethod
    def get_algorithm_data(cls, distribution_type: DistributionType,
                           component_number: int):
        cls.__cache_lock.acquire()
        key = (distribution_type, component_number)
        if key in cls.__cache:
            data = cls.__cache[key]
        else:
            data = AlgorithmData(distribution_type, component_number)
            cls.__cache[key] = data
        cls.__cache_lock.release()
        return data

    def process_params(self, fitted_params: Iterable, x_offset: float) -> Tuple[Tuple[Tuple, float]]:
        params_copy = np.array(fitted_params)
        param_count = get_param_count(self.distribution_type)
        if self.distribution_type == DistributionType.Normal or self.distribution_type == DistributionType.GeneralWeibull:
            for i in range(self.component_number):
                params_copy[i*param_count] += x_offset
        return process_params(self.distribution_type, self.component_number, params_copy)

    def get_param_by_mean(self, mean_values: Iterable):
        return get_param_by_mean(self.distribution_type, self.component_number, mean_values)


if __name__ == "__main__":
    # test the generating speed of algorithm data
    import time
    import sys
    start_uncached = time.time()
    data_list_uncached = []
    for i in range(10000):
        for component_number in range(3, 11):
            data = AlgorithmData(DistributionType.GeneralWeibull, component_number)
            data_list_uncached.append(data)
    end_uncached = time.time()
    print("Uncached time spent:", end_uncached-start_uncached, "s")

    start_cached = time.time()
    data_list_cached = []
    for i in range(10000):
        for component_number in range(3, 11):
            data = AlgorithmData.get_algorithm_data(DistributionType.GeneralWeibull, component_number)
            data_list_cached.append(data)
    end_cached = time.time()
    print("Cached time spent:", end_cached-start_cached, "s")



__all__ = ["PCAPanel"]

import csv
import logging
import os

import numpy as np
import xlsxwriter
import xlwt
from PySide2.QtCharts import QtCharts
from PySide2.QtCore import Qt
from PySide2.QtGui import QRegExpValidator
from PySide2.QtWidgets import (QCheckBox, QFileDialog, QGridLayout, QLabel,
                               QLineEdit, QMessageBox, QPushButton, QWidget)
from sklearn.decomposition import PCA

from QGrain.models.SampleDataset import SampleDataset
from QGrain.ui.Canvas import Canvas


class PCAPanel(Canvas):
    logger = logging.getLogger("root.ui.PCAPanel")
    gui_logger = logging.getLogger("GUI")

    def __init__(self, parent=None, is_dark=True):
        super().__init__("PCAPanel", parent=parent)
        self.init_chart()
        self.set_theme_mode(is_dark)
        self.setup_chart_style()
        self.init_ui()
        self.chart.legend().detachFromChart()
        self.chart.legend().setPos(100.0, 60.0)

        self.file_dialog = QFileDialog(self)
        self.msg_box = QMessageBox(self)
        self.msg_box.setWindowFlags(Qt.Drawer)
        self.retry_msg_box = QMessageBox(self)
        self.retry_msg_box.addButton(QMessageBox.StandardButton.Retry)
        self.retry_msg_box.addButton(QMessageBox.StandardButton.Ok)
        self.retry_msg_box.setDefaultButton(QMessageBox.StandardButton.Retry)
        self.retry_msg_box.setWindowFlags(Qt.Drawer)

        self.plot_data_items = []
        self.dataset = None
        self.transformed = None
        self.inverse_transformed = None


    def init_ui(self):
        # use anotehr widget to pack other controls
        self.control_container = QWidget(self)
        self.control_layout = QGridLayout(self.control_container)
        self.control_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.addWidget(self.control_container, 1, 0)
        # prepare controls
        self.component_number_validator = QRegExpValidator(r"^[1-9]\d*$") # >= 1
        self.fraction_validator = QRegExpValidator(r"^(0(\.\d{1,4})?|1(\.0{1,4})?)$") # 0.0000 - 1.0000
        self.assign_component_number_checkbox = QCheckBox(self.tr("Assign Component Number"))
        self.assign_component_number_checkbox.setCheckState(Qt.Unchecked)
        self.checkbox_state_label = QLabel(self.tr("Least Information Fraction:"))
        self.param_edit = QLineEdit()
        self.param_edit.setValidator(self.fraction_validator)
        self.param_edit.setText("0.9")
        self.perform_button = QPushButton(self.tr("Perform"))
        self.save_button = QPushButton(self.tr("Save"))
        self.control_layout.addWidget(self.assign_component_number_checkbox, 0, 0)
        self.control_layout.addWidget(self.checkbox_state_label, 0, 1)
        self.control_layout.addWidget(self.param_edit, 0, 2)
        self.control_layout.addWidget(self.perform_button, 0, 3)
        self.control_layout.addWidget(self.save_button, 0, 4)
        # connect
        self.assign_component_number_checkbox.stateChanged.connect(self.on_assign_checkbox_changed)
        self.perform_button.clicked.connect(self.on_perform_clicked)
        self.save_button.clicked.connect(self.on_save_clicked)

    def init_chart(self):
        # init axes
        self.axis_x = QtCharts.QValueAxis()
        self.axis_x.setLabelFormat("%i")
        self.chart.addAxis(self.axis_x, Qt.AlignBottom)
        self.axis_y = QtCharts.QValueAxis()
        self.chart.addAxis(self.axis_y, Qt.AlignLeft)
        # set title
        self.chart.setTitle(self.tr("PCA Canvas"))
        # set labels
        self.axis_x.setTitleText(self.tr("Sample Index"))
        self.axis_y.setTitleText(self.tr("Transformed"))
        # use demo to let it perform normal
        self.show_demo(self.axis_x, self.axis_y)

    def show_message(self, title: str, message: str):
        self.msg_box.setWindowTitle(title)
        self.msg_box.setText(message)
        self.msg_box.exec_()

    def show_info(self, message: str):
        self.show_message(self.tr("Info"), message)

    def show_warning(self, message: str):
        self.show_message(self.tr("Warning"), message)

    def show_error(self, message: str):
        self.show_message(self.tr("Error"), message)

    def on_assign_checkbox_changed(self, state: Qt.CheckState):
        if state == Qt.Checked:
            self.checkbox_state_label.setText(self.tr("Component Number:"))
            self.param_edit.setValidator(self.component_number_validator)
            self.param_edit.setText("1")
        else:
            self.checkbox_state_label.setText(self.tr("Least Information Fraction:"))
            self.param_edit.setValidator(self.fraction_validator)
            self.param_edit.setText("0.9")

    def get_param_value(self):
        if self.param_edit.text() == "":
            self.show_warning(self.tr("The component number / least information fraction is necessary."))
            return None
        if self.assign_component_number_checkbox.checkState() == Qt.Checked:
            return int(self.param_edit.text())
        else:
            return float(self.param_edit.text())

    def on_data_loaded(self, dataset: SampleDataset):
        self.dataset = dataset
        self.on_perform_clicked()

    def get_data_matrix(self):
        if self.dataset is None:
            self.show_warning(self.tr("Please load data first."))
            return None
        matrix = []
        for sample in self.dataset.samples:
            matrix.append(sample.distribution)
        return np.array(matrix)

    def on_perform_clicked(self):
        # necessary to stop
        self.stop_demo()
        n_components = self.get_param_value()
        if n_components is None:
            return
        matrix = self.get_data_matrix()
        if matrix is None:
            return
        self.logger.debug("Start to perform PCA algorithm, n_components is [%s].", n_components)
        pca = PCA(n_components=n_components)
        transformed = pca.fit_transform(matrix)
        sample_number, dimension_number = transformed.shape
        x = np.array(range(1, sample_number+1))
        # clear
        self.chart.removeAllSeries()
        for i in range(dimension_number):
            component_name = "C{0} ({1:0.2%})".format(i+1, pca.explained_variance_ratio_[i])
            series = QtCharts.QLineSeries()
            series.setName(component_name)
            series.replace(self.to_points(x, transformed[:, i]))
            self.chart.addSeries(series)
            series.attachAxis(self.axis_x)
            series.attachAxis(self.axis_y)
        # update the size of legend
        self.chart.legend().setMinimumSize(120.0, 30*(2+dimension_number))
        # reset the range of axes
        self.axis_x.setRange(x[0], x[-1])
        self.axis_y.setRange(round(np.min(transformed)*1.2, 2), round(np.max(transformed)*1.2, 2))

        self.transformed = transformed
        # split the pca components
        def get_split_distributions(index):
            processed = np.zeros_like(transformed)
            processed[:, index] = transformed[:, index]
            return pca.inverse_transform(processed)
        self.inverse_transformed = [get_split_distributions(i) for i in range(dimension_number)]

        self.logger.debug("PCA algorithm performed.")

    def on_save_clicked(self):
        if self.transformed is None or self.inverse_transformed is None:
            self.show_warning(self.tr("The PCA algorithm has not been performed."))
            return
        filename, type_str = self.file_dialog.getSaveFileName(None, self.tr("Save Recorded Data"), None, "Excel (*.xlsx);;97-2003 Excel (*.xls);;CSV (*.csv)")

        if filename is None or filename == "":
            self.logger.info("The path is None or empty, ignored.")
            return
        if os.path.exists(filename):
            self.logger.warning("This file has existed and will be replaced. Filename: %s.", filename)
        self.logger.info("File path to save is [%s].", filename)
        if ".xlsx" in type_str:
            self.save_as_xlsx(filename)
        elif "97-2003" in type_str:
            self.save_as_xls(filename)
        elif ".csv" in type_str:
            self.save_as_csv(filename)
        else:
            raise NotImplementedError(type_str)
        self.logger.info("PCA result has been saved to [%s].", filename)
        self.show_info(self.tr("PCA result has been saved to:\n[%s].") % filename)

    def save_as_csv(self, filename: str):
        with open(filename, "w", newline="", encoding="utf-8") as f:
            w = csv.writer(f)
            sample_number, dimension_number = self.transformed.shape
            headers = ["Sample Name"] + ["Component {0}".format(i+1) for i in range(dimension_number)]
            w.writerow(headers)
            for i in range(sample_number):
                w.writerow([self.dataset.samples[i].name] + list(self.transformed[i]))

    def save_as_xls(self, filename: str):
        sample_number, dimension_number = self.transformed.shape
        workbook = xlwt.Workbook()
        sheet = workbook.add_sheet("PCA Result")
        sheet.write(0, 0, "Sample Name")
        for i in range(dimension_number):
            sheet.write(0, i+1, "Component {0}".format(i+1))
        for sample_index in range(sample_number):
            sheet.write(sample_index+1, 0, self.dataset.samples[sample_index].name)
            for dimension_index in range(dimension_number):
                sheet.write(sample_index+1, dimension_index+1, self.transformed[sample_index, dimension_index])

        for i, split_distribution in enumerate(self.inverse_transformed):
            component_sheet = workbook.add_sheet("Component"+str(i+1))
            component_sheet.write(0, 0, "Sample Name")
            classes_number = len(self.dataset.classes)
            for j in range(classes_number):
                component_sheet.write(0, j+1, self.dataset.classes[j])
            for sample_index in range(sample_number):
                component_sheet.write(sample_index+1, 0, self.dataset.samples[sample_index].name)
                for classes_index in range(classes_number):
                    component_sheet.write(sample_index+1, classes_index+1, split_distribution[sample_index, classes_index])

        workbook.save(filename)

    def save_as_xlsx(self, filename: str):
        sample_number, dimension_number = self.transformed.shape
        workbook = xlsxwriter.Workbook(filename)
        sheet = workbook.add_worksheet("PCA Result")
        sheet.write(0, 0, "Sample Name")
        for i in range(dimension_number):
            sheet.write(0, i+1, "Component {0}".format(i+1))
        for sample_index in range(sample_number):
            sheet.write(sample_index+1, 0, self.dataset.samples[sample_index].name)
            for dimension_index in range(dimension_number):
                sheet.write(sample_index+1, dimension_index+1, self.transformed[sample_index, dimension_index])

        for i, split_distribution in enumerate(self.inverse_transformed):
            component_sheet = workbook.add_worksheet("Component"+str(i+1))
            component_sheet.write(0, 0, "Sample Name")
            classes_number = len(self.dataset.classes)
            for j in range(classes_number):
                component_sheet.write(0, j+1, self.dataset.classes[j])
            for sample_index in range(sample_number):
                component_sheet.write(sample_index+1, 0, self.dataset.samples[sample_index].name)
                for classes_index in range(classes_number):
                    component_sheet.write(sample_index+1, classes_index+1, split_distribution[sample_index, classes_index])

        workbook.close()


if __name__ == "__main__":
    import sys
    from PySide2.QtWidgets import QApplication
    app = QApplication(sys.argv)
    panel = PCAPanel(is_dark=False)
    panel.chart.legend().hide()
    panel.show()
    sys.exit(app.exec_())



__all__ = ["AppSetting"]

import logging
import os

from PySide2.QtCore import QSettings, Qt, Signal
from PySide2.QtWidgets import QComboBox, QGridLayout, QLabel, QWidget

QGRAIN_ROOT_PATH = os.path.dirname(os.path.dirname(__file__))


class AppSetting(QWidget):
    logger = logging.getLogger("root.ui.AppSetting")
    gui_logger = logging.getLogger("GUI")
    def __init__(self):
        super().__init__()
        self.language_options = [("简体中文", "zh_CN"),
                                 ("English", "en")]
        self.theme_options = [("Aqua", "Aqua"),
                              ("Elegant Dark", "ElegantDark"),
                              ("Material Dark", "MaterialDark"),
                              ("Ubuntu", "Ubuntu")]
        self.settings = QSettings(os.path.join(QGRAIN_ROOT_PATH, "settings", "QGrain.ini"), QSettings.Format.IniFormat)
        self.settings.beginGroup("app")
        self.setAttribute(Qt.WA_StyledBackground, True)
        self.init_ui()

    def init_ui(self):
        self.main_layout = QGridLayout(self)
        self.title_label = QLabel(self.tr("App Settings:"))
        self.title_label.setStyleSheet("QLabel {font: bold;}")
        self.main_layout.addWidget(self.title_label, 0, 0)
        self.language_label = QLabel(self.tr("Language"))
        self.language_combo_box = QComboBox()
        self.language_combo_box.addItems([name for name, enum_value in self.language_options])
        self.language_combo_box.setMaxVisibleItems(5)
        self.theme_label = QLabel(self.tr("Theme"))
        self.theme_combo_box = QComboBox()
        self.theme_combo_box.addItems([name for name, enum_value in self.theme_options])
        self.main_layout.addWidget(self.language_label, 1, 0)
        self.main_layout.addWidget(self.language_combo_box, 1, 1)
        self.main_layout.addWidget(self.theme_label, 2, 0)
        self.main_layout.addWidget(self.theme_combo_box, 2, 1)

        self.language_combo_box.currentIndexChanged.connect(self.on_language_changed)
        self.theme_combo_box.currentIndexChanged.connect(self.on_theme_changed)
        self.restore()

    def on_language_changed(self, index: int):
        _, lang = self.language_options[index]
        self.settings.setValue("language", lang)
        self.logger.info("Language has been changed to [%s].", lang)

    def on_theme_changed(self, index: int):
        _, theme = self.theme_options[index]
        self.settings.setValue("theme", theme)
        self.logger.info("Theme has been changed to [%s].", theme)

    def restore(self):
        language = self.settings.value("language", defaultValue="en", type=str)
        for i, (_, language_value) in enumerate(self.language_options):
            if language == language_value:
                self.language_combo_box.setCurrentIndex(i)
                break
        theme = self.settings.value("theme", defaultValue="MaterialDark", type=str)
        for i, (_, theme_value) in enumerate(self.theme_options):
            if theme == theme_value:
                self.theme_combo_box.setCurrentIndex(i)
                break



__all__ = ["MultiProcessingResolver"]

import logging
import time
from multiprocessing import Pool, cpu_count
from typing import List

from PySide2.QtCore import QObject, Signal

from QGrain.algorithms import DistributionType
from QGrain.models.SampleDataset import SampleDataset
from QGrain.resolvers.HeadlessResolver import FittingTask, HeadlessResolver


def run_task(task):
    global resolver
    results = resolver.execute_task(task)
    return results

def setup_process(*args):
    global resolver
    resolver = HeadlessResolver()
    for distribution_type, component_number in args:
        resolver.distribution_type = distribution_type
        resolver.component_number = component_number

class MultiProcessingResolver(QObject):
    sigTaskStateUpdated = Signal(tuple)
    sigTaskFinished = Signal(list, list)
    sigTaskInitialized = Signal(list)
    logger = logging.getLogger(name="root.resolvers.MultiProcessingResolver")
    STATE_CHECK_TIME_INTERVAL = 0.1

    def __init__(self):
        super().__init__()
        self.component_number = 3
        self.distribution_type = DistributionType.GeneralWeibull
        self.algorithm_settings = None

        self.grain_size_data = None # type: SampleDataset
        self.tasks = None # type: List[FittingTask]


    def on_component_number_changed(self, component_number: int):
        self.component_number = component_number
        self.logger.info("Component number has been changed to [%d].", self.component_number)

    def on_distribution_type_changed(self, distribution_type: DistributionType):
        self.distribution_type = DistributionType(distribution_type)
        self.logger.info("Distribution type has been changed to [%s].", self.distribution_type)

    def on_algorithm_settings_changed(self, settings: dict):
        self.algorithm_settings = settings
        self.logger.info("Algorithm settings have been changed to [%s].", settings)

    def on_data_loaded(self, data: SampleDataset):
        if data is None:
            return
        elif not data.has_data:
            return

        self.grain_size_data = data

    def init_tasks(self):
        tasks = []
        for sample in self.grain_size_data.samples:
            task = FittingTask(
                sample,
                component_number=self.component_number,
                distribution_type=self.distribution_type,
                algorithm_settings=self.algorithm_settings)
            tasks.append(task)
        self.tasks = tasks
        self.sigTaskInitialized.emit(tasks)

    def execute_tasks(self):
        if self.grain_size_data is None:
            return
        self.init_tasks()

        async_results = [(task.sample.uuid, self.pool.apply_async(run_task, args=(task,))) for task in self.tasks]

        while True:
            time.sleep(self.STATE_CHECK_TIME_INTERVAL)
            task_states = []
            for sample_id, result in async_results:
                task_states.append((sample_id, result.ready()))
            all_ready = True
            for sample_id, ready in task_states:
                if not ready:
                    all_ready = False
                    break
            self.sigTaskStateUpdated.emit(task_states)
            if all_ready:
                break

        succeeded_results = []
        failed_tasks = []
        for sample_id, r in async_results:
            flag, task, fitting_result = r.get()
            if flag:
                succeeded_results.append(fitting_result)
            else:
                failed_tasks.append(task)

        self.sigTaskFinished.emit(succeeded_results, failed_tasks)


    def setup_all(self):
        suggested_params = [(DistributionType.GeneralWeibull, 1), (DistributionType.GeneralWeibull, 2),
                            (DistributionType.GeneralWeibull, 3), (DistributionType.GeneralWeibull, 4)]
        self.pool = Pool(cpu_count(), setup_process, suggested_params)

    def cleanup_all(self):
        self.pool.terminate()
        self.pool.join()



__all__ = ["AlgorithmSetting"]

import logging
import os

from PySide2.QtCore import QSettings, Qt, Signal
from PySide2.QtGui import QDoubleValidator, QIntValidator, QValidator
from PySide2.QtWidgets import (QGridLayout, QLabel, QLineEdit, QMessageBox,
                               QWidget)

QGRAIN_ROOT_PATH = os.path.dirname(os.path.dirname(__file__))


class AlgorithmSetting(QWidget):
    sigAlgorithmSettingChanged = Signal(dict)
    logger = logging.getLogger("root.ui.AlgorithmSetting")
    gui_logger = logging.getLogger("GUI")
    def __init__(self):
        super().__init__()
        self.msg_box = QMessageBox(self)
        self.msg_box.setWindowFlags(Qt.Drawer)
        self.settings = QSettings(os.path.join(QGRAIN_ROOT_PATH, "settings", "QGrain.ini"), QSettings.Format.IniFormat)
        self.settings.beginGroup("algorithm")
        self.setAttribute(Qt.WA_StyledBackground, True)
        self.init_ui()

    def init_ui(self):
        self.main_layout = QGridLayout(self)
        self.int_validator = QIntValidator()
        self.int_validator.setBottom(0)
        self.double_validator = QDoubleValidator()
        self.double_validator.setBottom(0.0)

        self.title_label = QLabel(self.tr("Algorithm Settings:"))
        self.title_label.setStyleSheet("QLabel {font: bold;}")
        self.main_layout.addWidget(self.title_label, 0, 0)

        self.global_maxiter_label = QLabel(self.tr("Global Optimization Max Iteration"))
        self.global_maxiter_label.setToolTip(self.tr("Max iteration number of global optimization.\nIf the global optimization iteration has reached the max number, fitting process will stop."))
        self.main_layout.addWidget(self.global_maxiter_label, 1, 0)
        self.global_maxiter_edit = QLineEdit()
        self.global_maxiter_edit.setValidator(self.int_validator)
        self.main_layout.addWidget(self.global_maxiter_edit, 1, 1)

        self.global_success_iter_label = QLabel(self.tr("Global Optimization Success Iteration"))
        self.global_success_iter_label.setToolTip(self.tr("It's one of the terminal conditions of global optimization.\nIt means the iteration number of reaching the same minimum."))
        self.main_layout.addWidget(self.global_success_iter_label, 2, 0)
        self.global_success_iter_edit = QLineEdit()
        self.global_success_iter_edit.setValidator(self.int_validator)
        self.main_layout.addWidget(self.global_success_iter_edit, 2, 1)

        self.global_stepsize_label = QLabel(self.tr("Global Optimization Stepsize"))
        self.global_stepsize_label.setToolTip(self.tr("The stepsize of searching global minimum.\nGreater stepsize will jump out the local minimum easier but may miss the global minimum."))
        self.main_layout.addWidget(self.global_stepsize_label, 3, 0)
        self.global_stepsize_edit = QLineEdit()
        self.global_stepsize_edit.setValidator(self.double_validator)
        self.main_layout.addWidget(self.global_stepsize_edit, 3, 1)

        self.minimizer_tolerance_level_label = QLabel(self.tr("Global Minimizer Tolerance Level"))
        self.minimizer_tolerance_level_label.setToolTip(self.tr("The tolerance level of the minimizer of global optimization.\nTolerance level means the accepted minimum variation (10 ^ -level) of the target function.\nIt controls the precision and speed of fitting.\nIt's recommended to use ralatively lower level in global optimization process but higher leverl in final fitting."))
        self.main_layout.addWidget(self.minimizer_tolerance_level_label, 4, 0)
        self.minimizer_tolerance_level_edit = QLineEdit()
        self.minimizer_tolerance_level_edit.setValidator(self.int_validator)
        self.main_layout.addWidget(self.minimizer_tolerance_level_edit, 4, 1)

        self.minimizer_maxiter_label = QLabel(self.tr("Global Minimizer Max Iteration"))
        self.minimizer_maxiter_label.setToolTip(self.tr("Max iteration number of the minimizer of global optimization."))
        self.main_layout.addWidget(self.minimizer_maxiter_label, 5, 0)
        self.minimizer_maxiter_edit = QLineEdit()
        self.minimizer_maxiter_edit.setValidator(self.int_validator)
        self.main_layout.addWidget(self.minimizer_maxiter_edit, 5, 1)

        self.final_tolerance_level_label = QLabel(self.tr("Final Fitting Tolerance Level"))
        self.final_tolerance_level_label.setToolTip(self.tr("The tolerance level of the minimizer of final fitting."))
        self.main_layout.addWidget(self.final_tolerance_level_label, 6, 0)
        self.final_tolerance_level_edit = QLineEdit()
        self.final_tolerance_level_edit.setValidator(self.int_validator)
        self.main_layout.addWidget(self.final_tolerance_level_edit, 6, 1)

        self.final_maxiter_label = QLabel(self.tr("Final Fitting Max Iteration"))
        self.final_maxiter_label.setToolTip(self.tr("Max iteration number of the minimizer of final fitting."))
        self.main_layout.addWidget(self.final_maxiter_label, 7, 0)
        self.final_maxiter_edit = QLineEdit()
        self.final_maxiter_edit.setValidator(self.int_validator)
        self.main_layout.addWidget(self.final_maxiter_edit, 7, 1)

        self.global_maxiter_edit.textChanged.connect(self.on_settings_changed)
        self.global_success_iter_edit.textChanged.connect(self.on_settings_changed)
        self.global_stepsize_edit.textChanged.connect(self.on_settings_changed)
        self.minimizer_tolerance_level_edit.textChanged.connect(self.on_settings_changed)
        self.minimizer_maxiter_edit.textChanged.connect(self.on_settings_changed)
        self.final_tolerance_level_edit.textChanged.connect(self.on_settings_changed)
        self.final_maxiter_edit.textChanged.connect(self.on_settings_changed)

        self.restore()

    def on_settings_changed(self):
        global_optimization_maxiter = self.global_maxiter_edit.text()
        global_optimization_success_iter = self.global_success_iter_edit.text()
        global_optimization_stepsize = self.global_stepsize_edit.text()
        minimizer_tolerance_level = self.minimizer_tolerance_level_edit.text()
        minimizer_maxiter = self.minimizer_maxiter_edit.text()
        final_tolerance_level = self.final_tolerance_level_edit.text()
        final_maxiter = self.final_maxiter_edit.text()

        try:
            signal_data = dict(
                global_optimization_maxiter=int(global_optimization_maxiter),
                global_optimization_success_iter=int(global_optimization_success_iter),
                global_optimization_stepsize=float(global_optimization_stepsize),
                minimizer_tolerance=10**(-int(minimizer_tolerance_level)),
                minimizer_maxiter=int(minimizer_maxiter),
                final_tolerance=10**(-int(final_tolerance_level)),
                final_maxiter=int(final_maxiter))

            self.sigAlgorithmSettingChanged.emit(signal_data)
        except ValueError:
            return

        self.settings.setValue("global_optimization_maxiter", global_optimization_maxiter)
        self.settings.setValue("global_optimization_success_iter", global_optimization_success_iter)
        self.settings.setValue("global_optimization_stepsize", global_optimization_stepsize)
        self.settings.setValue("minimizer_tolerance_level", minimizer_tolerance_level)
        self.settings.setValue("minimizer_maxiter", minimizer_maxiter)
        self.settings.setValue("final_tolerance_level", final_tolerance_level)
        self.settings.setValue("final_maxiter", final_maxiter)

    def restore(self):
        self.global_maxiter_edit.setText(self.settings.value("global_optimization_maxiter", defaultValue="100"))
        self.global_success_iter_edit.setText(self.settings.value("global_optimization_success_iter", defaultValue="3"))
        self.global_stepsize_edit.setText(self.settings.value("global_optimization_stepsize", defaultValue="1.0"))
        self.minimizer_tolerance_level_edit.setText(self.settings.value("minimizer_tolerance_level", defaultValue="8"))
        self.minimizer_maxiter_edit.setText(self.settings.value("minimizer_maxiter", defaultValue="500"))
        self.final_tolerance_level_edit.setText(self.settings.value("final_tolerance_level", defaultValue="100"))
        self.final_maxiter_edit.setText(self.settings.value("final_maxiter", defaultValue="1000"))



from setuptools import setup, find_packages
import os

setup(
    name="QGrain",
    version="0.2.7.1",
    description="QGrain is an esay to use tool that can unmix and analyse the multi-modal grain size distribution.",
    platforms="all",
    author="Yuming Liu",
    author_email="liuyuming@ieecas.cn",
    url="https://github.com/QGrain-Organization/QGrain",
    license="MIT",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'numpy>=1.18.1',
        'PySide2>=5.14.1',
        "scikit-learn>=0.22.1",
        "scipy>=1.4.1",
        "shiboken2>=5.14.1",
        "xlrd>=1.2.0",
        "XlsxWriter>=1.2.7",
        "xlwt>=1.3.0",
        "Pillow>=7.0.0",
        "opencv-python>=4.2.0.32"
    ],
    entry_points = {
        'console_scripts': ['qgrain=QGrain.main:main'],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)



__all__ = ["FittingTask", "FinalLocalOptimizationError",
           "GlobalOptimizationError",
           "HeadlessResolver"]

import numpy as np
from scipy.optimize import OptimizeResult

from QGrain.algorithms import DistributionType
from QGrain.models.FittingResult import FittingResult
from QGrain.models.SampleData import SampleData
from QGrain.resolvers.Resolver import Resolver


class FittingTask:
    def __init__(self, sample: SampleData,
                 distribution_type=DistributionType.GeneralWeibull,
                 component_number=3,
                 algorithm_settings=None):
        self.sample = sample
        self.component_number = component_number
        self.distribution_type = distribution_type
        self.algorithm_settings = algorithm_settings


class FinalLocalOptimizationError(Exception):
    """
    Raises when the final local optimization failed.
    """
    pass


class GlobalOptimizationError(Exception):
    """
    Raises when the global optimization failed.
    """
    pass


class HeadlessResolver(Resolver):
    def __init__(self):
        super().__init__()
        self.current_task = None # type: FittingTask
        self.current_result = None # type: FittingResult
        self.current_exception = None # type: Exception

    def on_fitting_succeeded(self, algorithm_result: OptimizeResult):
        result = self.get_fitting_result(algorithm_result.x)
        self.current_result = result

    def on_global_fitting_failed(self, algorithm_result: OptimizeResult):
        self.current_exception = GlobalOptimizationError(algorithm_result.message)

    def on_final_fitting_failed(self, algorithm_result: OptimizeResult):
        self.current_exception = FinalLocalOptimizationError(algorithm_result.message)

    def on_exception_raised_while_fitting(self, exception: Exception):
        self.current_exception

    def execute_task(self, task: FittingTask):
        self.current_task = task
        self.distribution_type = task.distribution_type
        self.component_number = task.component_number
        if task.algorithm_settings is not None:
            self.change_settings(**task.algorithm_settings)
        self.feed_data(task.sample)
        self.try_fit()
        if self.current_result is None:
            assert self.current_exception is not None
            return False, self.current_task, self.current_exception
        else:
            return True, self.current_task, self.current_result



__all__ = ["CancelError", "GUIResolver"]

import logging
import time

import numpy as np
from PySide2.QtCore import QMutex, QObject, Signal, Slot
from scipy.interpolate import interp1d
from scipy.optimize import OptimizeResult

from QGrain.algorithms import DistributionType
from QGrain.models.FittingResult import FittingResult
from QGrain.models.SampleData import SampleData
from QGrain.resolvers.Resolver import Resolver


class CancelError(Exception):
    pass


class GUIResolver(QObject, Resolver):
    sigFittingStarted = Signal()
    sigFittingFinished = Signal()
    sigFittingSucceeded = Signal(FittingResult)
    sigFittingFailed = Signal(str) # emit hint text
    logger = logging.getLogger(name="root.resolvers.GUIResolver")

    def __init__(self, inherit_params=True):
        super().__init__()
        Resolver.__init__(self)
        # settings
        self.expected_params = None
        self.inherit_params = inherit_params
        self.last_succeeded_params = None

        self.cancel_flag = False
        self.cancel_mutex = QMutex()

    def on_component_number_changed(self, component_number: int):
        self.component_number = component_number
        self.logger.info("Component Number has been changed to [%d].", self.component_number)

    def on_distribution_type_changed(self, distribution_type: DistributionType):
        self.distribution_type = DistributionType(distribution_type)
        self.logger.info("Distribution type has been changed to [%s].", self.distribution_type)
        # clear if type changed
        self.last_succeeded_params = None

    def on_inherit_params_changed(self, value: bool):
        self.inherit_params = value
        self.logger.info("Setting [%s] have been changed to [%s].", "inherit_params", value)

    def on_algorithm_settings_changed(self, settings: dict):
        self.change_settings(**settings)
        self.logger.info("Algorithm settings have been changed to [%s].", settings)

    def on_target_data_changed(self, sample: SampleData):
        self.logger.debug("Target data has been changed to [%s].", sample.name)
        self.feed_data(sample)

    def on_data_fed(self, sample_name: str):
        self.logger.debug("Sample [%s] has been fed.", sample_name)

    def on_data_not_prepared(self):
        self.sigFittingFailed.emit(self.tr("There is no valid data to fit."))
        self.logger.error("There is no valid data to fit.")

    def on_fitting_started(self):
        super().on_fitting_started()
        if self.inherit_params and \
            self.last_succeeded_params is not None and \
            len(self.last_succeeded_params) == len(self.algorithm_data.defaults):
            self.initial_guess = self.last_succeeded_params
        else:
            self.initial_guess = self.algorithm_data.defaults

        if self.expected_params is not None:
            self.initial_guess = self.expected_params

        self.sigFittingStarted.emit()
        self.logger.debug("Fitting progress started.")

    def on_fitting_finished(self):
        self.expected_params = None
        self.sigFittingFinished.emit()
        self.logger.debug("Fitting progress finished.")

    def on_global_fitting_failed(self, algorithm_result: OptimizeResult):
        self.sigFittingFailed.emit(self.tr("Fitting failed during global fitting progress."))
        self.logger.error("Fitting failed during global fitting progress. Details: [%s].", algorithm_result.message)

    def on_global_fitting_succeeded(self, algorithm_result: OptimizeResult):
        self.logger.debug("Global fitting progress succeeded.")

    def on_final_fitting_failed(self, algorithm_result: OptimizeResult):
        self.sigFittingFailed.emit(self.tr("Fitting failed during final fitting progress."))
        self.logger.error("Fitting failed during final fitting progress. Details: [%s].", algorithm_result.message)

    def on_exception_raised_while_fitting(self, exception: Exception):
        if type(exception) == CancelError:
            self.logger.info("The fitting progress was canceled by user.")
        else:
            self.sigFittingFailed.emit(self.tr("Unknown exception raise in fitting progress."))
            self.logger.exception("Unknown exception raise in fitting progress.", stack_info=True)

    def local_iteration_callback(self, fitted_params):
        self.cancel_mutex.lock()
        if self.cancel_flag:
            self.cancel_flag = False
            self.cancel_mutex.unlock()
            # use exception to stop the fitting progress
            # need to handle the exception in `on_exception_raised_while_fitting` func
            raise CancelError()
        else:
            self.cancel_mutex.unlock()
        super().local_iteration_callback(fitted_params)

    def global_iteration_callback(self, fitted_params, function_value, accept):
        pass

    def on_fitting_succeeded(self, algorithm_result: OptimizeResult):
        # record the succeeded params
        self.last_succeeded_params = algorithm_result.x
        self.logger.info("The epoch of fitting has finished, the fitted parameters are: [%s]", algorithm_result.x)
        self.sigFittingSucceeded.emit(self.get_fitting_result(algorithm_result.x))

    # call this func by another thread
    # so, lock is necessary
    def cancel(self):
        self.cancel_mutex.lock()
        self.cancel_flag = True
        self.cancel_mutex.unlock()

    def on_excepted_mean_value_changed(self, mean_values):
        if not self.data_prepared:
            return
        x_real_to_space = interp1d(self.real_x, self.fitting_space_x)
        converted_x = [x_real_to_space(mean).max() - self.x_offset for mean in mean_values]
        converted_x.sort()
        self.expected_params = self.algorithm_data.get_param_by_mean(converted_x)



__all__ = ["ComponentFittingResult", "FittingResult"]

import copy
from typing import Callable, Dict, Iterable, List, Tuple
from uuid import UUID, uuid4

import numpy as np
from scipy.interpolate import interp1d
from scipy.stats import kendalltau, pearsonr, spearmanr

from QGrain.algorithms import AlgorithmData, DistributionType


class ComponentFittingResult:
    """
    The class to record the fitting result of each component.

    All its attributes have been packed with read-only properties to avoid the modification by mistake.

    Use `ctor` to initialize the instance and treat it as immutable object.
    If you have sufficient reasons to modify its attributes, please use the `update` method of it.

    Attributes:
        real_x: The `numpy.ndarray` which represents the real (raw) values of x (i.e. grain size classes).
        fitting_space_x: The `numpy.ndarray` which represents the x values (e.g. bin numbers) used in fitting process.
        params: The paramters which were used to calculate the y array (i.e. distribution) and statistic values of each component.
        fraction: The fraction of this component.
        mean, median, mode, etc: The statistic values of this component.
    """
    def __init__(self, real_x: np.ndarray, fitting_space_x: np.ndarray,
                 algorithm_data: AlgorithmData,
                 params: Iterable[float], fraction: float):
        assert np.isreal(fraction) and fraction is not None
        # iteration may pass invalid fraction value
        # assert fraction >= 0.0 and fraction <= 1.0

        self.__real_x = real_x
        self.__fitting_space_x = fitting_space_x
        if np.any(np.isnan(params)):
            self.__fraction = np.nan
            self.__component_y = np.full_like(fitting_space_x, fill_value=np.nan)
            self.__mean = np.nan
            self.__median = np.nan
            self.__mode = np.nan
            self.__variance = np.nan
            self.__standard_deviation = np.nan
            self.__skewness = np.nan
            self.__kurtosis = np.nan
        else:
            self.update(algorithm_data, params, fraction)

    def update(self, algorithm_data: AlgorithmData, params: Iterable[float], fraction: float):
        self.__params = params
        self.__fraction = fraction
        self.__component_y = algorithm_data.single_func(self.__fitting_space_x, *params) * fraction
        x_to_real = interp1d(self.__fitting_space_x, self.__real_x)
        try:
            self.__mean = x_to_real(algorithm_data.mean(*params)).max()
        except ValueError:
            self.__mean = np.nan
        try:
            self.__median = x_to_real(algorithm_data.median(*params)).max()
        except ValueError:
            self.__median = np.nan
        try:
            self.__mode = x_to_real(algorithm_data.mode(*params)).max()
        except ValueError:
            self.__mode = np.nan
        self.__variance = algorithm_data.variance(*params)
        self.__standard_deviation = algorithm_data.standard_deviation(*params)
        self.__skewness = algorithm_data.skewness(*params)
        self.__kurtosis = algorithm_data.kurtosis(*params)

    @property
    def params(self) -> Tuple[float]:
        return self.__params

    @property
    def fraction(self) -> float:
        return self.__fraction

    @property
    def component_y(self) -> np.ndarray:
        return self.__component_y

    @property
    def mean(self) -> float:
        return self.__mean

    @property
    def median(self) -> float:
        return self.__median

    @property
    def mode(self) -> float:
        return self.__mode

    @property
    def variance(self) -> float:
        return self.__variance

    @property
    def standard_deviation(self) -> float:
        return self.__standard_deviation

    @property
    def skewness(self) -> float:
        return self.__skewness

    @property
    def kurtosis(self) -> float:
        return self.__kurtosis

    @property
    def has_nan(self) -> bool:
        values_to_check = [self.fraction, self.mean,
                           self.median, self.mode,
                           self.variance, self.standard_deviation,
                           self.skewness, self.kurtosis]
        if np.any(np.isnan(self.component_y)) or np.any(np.isnan(values_to_check)):
            return True
        else:
            return False


class FittingResult:
    """
    The class to represent the fitting result of each sample.
    """
    def __init__(self, name: str, real_x: np.ndarray,
                 fitting_space_x: np.ndarray, bin_numbers: np.ndarray,
                 target_y: np.ndarray, algorithm_data: AlgorithmData,
                 fitted_params: Iterable[float], x_offset: float,
                 fitting_history: List[np.ndarray] = None):
        length = len(real_x)
        assert len(fitting_space_x) == length
        assert len(bin_numbers) == length
        assert len(target_y) == length

        # add uuid to manage data
        self.__uuid = uuid4()
        self.__name = name
        self.__real_x = real_x
        self.__fitting_space_x = fitting_space_x
        self.__bin_numbers = bin_numbers
        self.__x_offset = x_offset
        self.__target_y = target_y
        self.__distribution_type = algorithm_data.distribution_type
        self.__component_number = algorithm_data.component_number
        self.__param_count = algorithm_data.param_count
        self.__param_names = algorithm_data.param_names
        self.__fitting_history = [fitted_params] if fitting_history is None else fitting_history
        self.__components = [] # type: List[ComponentFittingResult]
        self.update(fitted_params)


    def update(self, fitted_params: Iterable[float]):
        algorithm_data = AlgorithmData.get_algorithm_data(self.distribution_type, self.component_number)
        if np.any(np.isnan(fitted_params)):
            self.__fitted_y = np.full_like(self.__fitting_space_x, fill_value=np.nan)
            self.__error_array = np.full_like(self.__fitting_space_x, fill_value=np.nan)
            self.__mean_squared_error = np.nan
            self.__pearson_r = (np.nan, np.nan)
            self.__kendall_tau = (np.nan, np.nan)
            self.__spearman_r = (np.nan, np.nan)
        else:
            self.__fitted_y = algorithm_data.mixed_func(self.__fitting_space_x - self.__x_offset, *fitted_params)
            # some test for fitting result
            self.__error_array = self.__target_y - self.__fitted_y
            self.__mean_squared_error = np.mean(np.square(self.__error_array))
            # https://scipy.github.io/devdocs/generated/scipy.stats.pearsonr.html
            self.__pearson_r = pearsonr(self.__target_y, self.__fitted_y)
            # https://scipy.github.io/devdocs/generated/scipy.stats.kendalltau.html
            self.__kendall_tau = kendalltau(self.__target_y, self.__fitted_y)
            # https://scipy.github.io/devdocs/generated/scipy.stats.spearmanr.html
            self.__spearman_r = spearmanr(self.__target_y, self.__fitted_y)

        processed_params = algorithm_data.process_params(fitted_params, self.__x_offset)
        if  len(self.__components) == 0:
            for params, fraction in processed_params:
                component_result = ComponentFittingResult(self.real_x, self.fitting_space_x, algorithm_data, params, fraction)
                self.__components.append(component_result)
        else:
            for component, (params, fraction) in zip(self.__components, processed_params):
                component.update(algorithm_data, params, fraction)
        # sort by mean values
        self.__components.sort(key=lambda component: component.mean)

    @property
    def uuid(self) -> UUID:
        return self.__uuid

    @property
    def name(self) -> str:
        return self.__name

    @property
    def real_x(self) -> np.ndarray:
        return self.__real_x

    @property
    def fitting_space_x(self) -> np.ndarray:
        return self.__fitting_space_x

    @property
    def bin_numbers(self) -> np.ndarray:
        return self.__bin_numbers

    @property
    def target_y(self) -> np.ndarray:
        return self.__target_y

    @property
    def distribution_type(self) -> DistributionType:
        return self.__distribution_type

    @property
    def component_number(self) -> int:
        return self.__component_number

    @property
    def param_count(self) -> int:
        return self.__param_count

    @property
    def param_names(self) -> Tuple[str]:
        return self.__param_names

    @property
    def components(self) -> Iterable[ComponentFittingResult]:
        return [component for component in self.__components]

    @property
    def fitted_y(self) -> np.ndarray:
        return self.__fitted_y

    @property
    def mean_squared_error(self) -> float:
        return self.__mean_squared_error

    @property
    def pearson_r(self) -> Tuple[float, float]:
        return self.__pearson_r

    @property
    def kendall_tau(self) -> Tuple[float, float]:
        return self.__kendall_tau

    @property
    def spearman_r(self) -> Tuple[float, float]:
        return self.__spearman_r

    @property
    def has_invalid_value(self) -> bool:
        if np.any(np.isnan(self.__fitted_y)):
            return True
        for component in self.__components:
            if component.has_nan:
                return True
        return False

    @property
    def history(self):
        copy_result = copy.deepcopy(self)
        for fitted_params in self.__fitting_history:
            copy_result.update(fitted_params)
            yield copy_result

    @property
    def iteration_number(self):
        return len(self.__fitting_history)



__all__ = ["FileType", "ValueNotNumberError", "CSVEncodingError", "DataLoader"]

import csv
from enum import Enum, unique
from typing import Iterable, Tuple

import numpy as np
from xlrd import XLRDError, open_workbook

from QGrain.models.DataLayoutSetting import DataLayoutError, DataLayoutSetting
from QGrain.models.SampleDataset import SampleDataset


class FileType:
    XLS = 0
    XLSX = 1
    CSV = 2

class ValueNotNumberError(Exception):
    "Raise while the value can not be converted to a number."
    pass

class CSVEncodingError(Exception):
    "Raise while the encoding of csv is not utf-8."
    pass


class DataLoader:
    """
    The class to load the grain size distributions from the loacl files.
    """
    def __init__(self):
        pass

    def try_load_data(self, filename: str, file_type: FileType, layout: DataLayoutSetting) -> SampleDataset:
        assert filename is not None
        assert filename != ""

        if file_type == FileType.CSV:
            return self.try_csv(filename, layout)
        elif file_type == FileType.XLS or \
                file_type == FileType.XLSX:
            return self.try_excel(filename, layout)
        else:
            raise NotImplementedError(file_type)

    def process_raw_data(self, raw_data: Iterable[Iterable],
                         assigned_layout: DataLayoutSetting = None,
                         replace_name_none="NONE", replace_name_empty="EMPTY") \
            -> Tuple[np.ndarray, Iterable[str], Iterable[np.ndarray]]:
        """
        Method to convert the raw data table to grain size classes, sample names and distributions.

        Args:
            raw_data: A 2-d `Iterable` object, e.g. List[List].
            assigned_layout: A `DataLayoutSetting` object, by default it will use the parameterless `ctor` of `DataLayoutSetting` to create one.
            replace_name_none: The `str` to replace the `None` value of sample names.
            replace_name_empty: Similar to above.

        Returns:
            A tuple that contains the classes array, sample names and distributions.

        Raises:
            DataLayoutError: If the `classes_row` or `sample_name_column` of the layout setting is out of range.
            ValueNotNumberError: If the value of classes or distributions can not be converted to real number.
        """
        if assigned_layout is None:
            layout = DataLayoutSetting()
        else:
            layout = assigned_layout
        try:
            classes = np.array(raw_data[layout.classes_row][layout.distribution_start_column:], dtype=np.float64)
            names = []
            distributions = []
            for row_values in raw_data[layout.distribution_start_row:]:
                sample_name = row_values[layout.sample_name_column]
                if sample_name is None:
                    sample_name = replace_name_none
                # users may use pure number as the sample name
                elif type(sample_name) != str:
                    sample_name = str(sample_name)
                elif sample_name == "":
                    sample_name = replace_name_empty
                distribution = np.array(row_values[layout.distribution_start_column:], dtype=np.float64)
                names.append(sample_name)
                distributions.append(distribution)

            return classes, names, distributions
        except IndexError as e:
            raise DataLayoutError("The data layout setting does not match the data file.") from e
        except ValueError as e:
            raise ValueNotNumberError("Some value can not be converted to real number, check the data file and layout setting.") from e

    def try_csv(self, filename: str, layout: DataLayoutSetting) -> SampleDataset:
        try:
            with open(filename, encoding="utf-8") as f:
                r = csv.reader(f)
                raw_data = [row for row in r]
                classes, names, distributions = self.process_raw_data(raw_data, layout)
                dataset = SampleDataset()
                dataset.add_batch(classes, names, distributions)
                return dataset
        except UnicodeDecodeError as e:
            raise CSVEncodingError("The encoding of csv file must be utf-8.") from e

    def try_excel(self, filename: str, layout: DataLayoutSetting) -> SampleDataset:
        sheet = open_workbook(filename).sheet_by_index(0)
        raw_data = []
        for i in range(sheet.nrows):
            raw_data.append(sheet.row_values(i))
        classes, names, distributions = self.process_raw_data(raw_data, layout)
        dataset = SampleDataset()
        dataset.add_batch(classes, names, distributions)
        return dataset
